;*******************************************************************************
;*
;*	Titel:	Unterprogramm zur Darstellung eines 16-Bit-Registers auf einer
;*		7-Segmentanzeige (SR_NUMDISPLAY)
;*
;*	Datum:	15/10/08	Version:	0.2	
;*	Autor:	WEIGU
;*
;*	Informationen zur Beschaltung:
;*	Prozessor:	ATmega32	Quarzfrequenz:	16MHz 
;*	Eingaenge:
;*	Ausgaenge:	Pin 0-6 (a-g) am SegPort
;*			Pin 0-3 (digit 0-3) am DigPort
;*
;*      !! alle 8 Bit von SegPort werden angesprochen. Bit 7 SegPort ist !!
;*      !! also nicht frei verfuegbar                                   !!
;*
;*	Informationen zur Funktionsweise:
;*	
;*	Der Inhalt des Doppelregisters Y wird waehrend einer Sekunde am Display
;*	ausgegeben.
;*	Im Hauptprogramm muessen beide Ports (DigPort, SegPort) initialisiert
;*	werden und die Zeitschleifenbibliothek muss eingebunden sein.
;*	Die Dekodiertabelle befindet sich im Programmspeicher.
;*	Die Hexstelle addiert zur Anfangsadresse dient als Adresszeiger.
;*
;*******************************************************************************

;-------------------------------------------------------------------------------
;	Unterprogramm zur Darstellung des Doppelreg. Y auf dem 7-Seg.-Display
;-------------------------------------------------------------------------------
;Verwendete Register:   r16,r17 ;Reg. 16 u. 17 dienen als Zwischenspeicher
;                       r18     ;Register 18 wird als Zaehler verwendet

NDISP:	push    r16             ;alle verwendeten Register (inkl. SREG) retten
        in      r16,SREG
        push    r16
        push    r17
        push    r18
        push	YL
        push	YH
        push    ZL
        push    ZH
        
        ;Zaehler initialisieren
        ldi     r18,50         ;Zaehler mit 250 initialisieren (250*4*1ms=1s)
NDISP0:	;erste Stelle anzeigen
        ldi     ZL,LOW(NDISPT*2)  ;Adresszeiger mit der Adresse der Tab.*2
        ldi     ZH,HIGH(NDISPT*2) ;initialisieren (Worte statt Bytes)
        mov     r16,YL          ;Adresszeiger errechnen
        andi    r16,0x0F        ;Maskieren des untersten Nibble von YL
        clr     r17             ;Zwischenspeicher = 0
        add     ZL,r16          ;Addiere Zaehlerstand zur Basisadresse
        adc     ZH,r17          ;ZH erhoehen falls Carry!
        cbi	PORTB,5         ;Erste Stelle einschalten
        in      r17,PINC        ;Zeichen am SegPort ausgeben (2 versch. Ports)
        andi    r17,0xF0        
        lpm     r16,Z           ;Zeichen im Flash abholen
        andi    r16,0x0F
        or      r16,r17
        out	PORTC,r16
        lpm     r16,Z           ;Zeichen im Flash abholen
        andi    r16,0x10        ;(Komplizierte Vorgehensweise um PD1 (ISR)
        breq    NDIS13          ;nicht zu stoeren)
        sbi     PORTD,4
NDIS11: lpm     r16,Z           ;Zeichen im Flash abholen
        andi    r16,0x20
        breq    NDIS14
        sbi     PORTD,5
NDIS12: lpm     r16,Z           ;Zeichen im Flash abholen
        andi    r16,0x40
        breq    NDIS15
        sbi     PORTD,6
        rjmp    NDIS16
NDIS13: cbi     PORTD,4
        rjmp    NDIS11
NDIS14: cbi     PORTD,5
        rjmp    NDIS12
NDIS15: cbi     PORTD,6
NDIS16: lds     Tmp1,ALFLAG
        cpi     Tmp1,1
        brne    NDIS17    
        sbi     PORTD,7         ;Trennpunkt ein
NDIS17: rcall	W1ms		;Warte 1ms
        cbi     PORTD,7         ;Trennpunkt aus
	sbi     PORTB,5         ;Erste Stelle ausschalten
        lds     Tmp1,ESFLAG     ;Stromsparmodus?
        cpi     Tmp1,1
        brne    NDISP2
        rcall   W1ms            ;Warte 1ms
        rcall   W1ms            ;Warte 1ms
        rcall   W1ms            ;Warte 1ms

	;zweite Stelle anzeigen
NDISP2:	ldi     ZL,LOW(NDISPT*2)  ;Adresszeiger mit der Adresse der Tab.*2
        ldi     ZH,HIGH(NDISPT*2) ;initialisieren (Worte statt Bytes)
        mov     r16,YL          ;Adresszeiger errechnen
        andi    r16,0xF0        ;Maskieren des obersten Nibble von YL
        swap    r16             ;Vertausche beide Nibble        
        clr     r17             ;Zwischenspeicher = 0
        add     ZL,r16          ;Addiere Zaehlerstand zur Basisadresse
        adc     ZH,r17          ;ZH erhoehen falls Carry!
        cbi     PORTB,4         ;Zweite Stelle einschalten
        in      r17,PINC        ;Zeichen am SegPort ausgeben (2 versch. Ports)
        andi    r17,0xF0
        lpm     r16,Z           ;Zeichen im Flash abholen
        andi    r16,0x0F
        or      r16,r17
        out     PORTC,r16
        in      r17,PIND
        andi    r17,0x8F
        lpm     r16,Z           ;Zeichen im Flash abholen
        andi    r16,0x10
        breq    NDIS23
        sbi     PORTD,4
NDIS21: lpm     r16,Z           ;Zeichen im Flash abholen
        andi    r16,0x20
        breq    NDIS24
        sbi     PORTD,5
NDIS22: lpm     r16,Z           ;Zeichen im Flash abholen
        andi    r16,0x40
        breq    NDIS25
        sbi     PORTD,6
        rjmp    NDIS26
NDIS23: cbi     PORTD,4
        rjmp    NDIS21
NDIS24: cbi     PORTD,5
        rjmp    NDIS22
NDIS25: cbi     PORTD,6
NDIS26: rcall   W1ms            ;Warte 1ms
        sbi     PORTB,4         ;Zweite Stelle ausschalten
        lds     Tmp1,ESFLAG     ;Stromsparmodus?
        cpi     Tmp1,1
        brne    NDISP3
        rcall   W1ms            ;Warte 1ms
        rcall   W1ms            ;Warte 1ms
        rcall   W1ms            ;Warte 1ms


	;dritte Stelle anzeigen
NDISP3: ldi     ZL,LOW(NDISPT*2)  ;Adresszeiger mit der Adresse der Tab.*2
        ldi     ZH,HIGH(NDISPT*2) ;initialisieren (Worte statt Bytes)
        mov     r16,YH          ;Adresszeiger errechnen
        andi    r16,0x0F        ;Maskieren des untersten Nibble von YH
        clr     r17             ;Zwischenspeicher = 0
        add     ZL,r16          ;Addiere Zaehlerstand zur Basisadresse
        adc     ZH,r17          ;ZH erhoehen falls Carry!
        cbi	PORTB,3 	;Dritte Stelle einschalten
        in      r17,PINC        ;Zeichen am SegPort ausgeben (2 versch. Ports)
        andi    r17,0xF0
        lpm     r16,Z           ;Zeichen im Flash abholen
        andi    r16,0x0F
        or      r16,r17
        out     PORTC,r16
        in      r17,PIND
        andi    r17,0x8F
        lpm     r16,Z           ;Zeichen im Flash abholen
        andi    r16,0x10
        breq    NDIS33
        sbi     PORTD,4
NDIS31: lpm     r16,Z           ;Zeichen im Flash abholen
        andi    r16,0x20
        breq    NDIS34
        sbi     PORTD,5
NDIS32: lpm     r16,Z           ;Zeichen im Flash abholen
        andi    r16,0x40
        breq    NDIS35
        sbi     PORTD,6
        rjmp    NDIS36
NDIS33: cbi     PORTD,4
        rjmp    NDIS31
NDIS34: cbi     PORTD,5
        rjmp    NDIS32
NDIS35: cbi     PORTD,6
NDIS36: sbi     PORTD,7         ;Trennpunkt ein
        rcall   W1ms            ;Warte 1ms
        cbi     PORTD,7         ;Trennpunkt aus
	sbi     PORTB,3         ;Dritte Stelle ausschalten
        lds     Tmp1,ESFLAG     ;Stromsparmodus?
        cpi     Tmp1,1
        brne    NDISP4
        rcall   W1ms            ;Warte 1ms
        rcall   W1ms            ;Warte 1ms
        rcall   W1ms            ;Warte 1ms


	;vierte Stelle anzeigen
 
NDISP4: ldi     ZL,LOW(NDISPT*2)  ;Adresszeiger mit der Adresse der Tab.*2
        ldi     ZH,HIGH(NDISPT*2) ;initialisieren (Worte statt Bytes)
        mov     r16,YH          ;Adresszeiger errechnen
        andi    r16,0xF0        ;Maskieren des obersten Nibble von YH
        swap    r16             ;Vertausche beide Nibble        
        clr     r17             ;Zwischenspeicher = 0
        add     ZL,r16          ;Addiere Zaehlerstand zur Basisadresse
        adc     ZH,r17          ;ZH erhoehen falls Carry!
        cbi	PORTB,2         ;Vierte Stelle einschalten
        in      r17,PINC        ;Zeichen am SegPort ausgeben (2 versch. Ports)
        andi    r17,0xF0
        lpm     r16,Z           ;Zeichen im Flash abholen
        andi    r16,0x0F
        or      r16,r17
        out     PORTC,r16
        in      r17,PIND
        andi    r17,0x8F
        lpm     r16,Z           ;Zeichen im Flash abholen
        andi    r16,0x10
        breq    NDIS43
        sbi     PORTD,4
NDIS41: lpm     r16,Z           ;Zeichen im Flash abholen
        andi    r16,0x20
        breq    NDIS44
        sbi     PORTD,5
NDIS42: lpm     r16,Z           ;Zeichen im Flash abholen
        andi    r16,0x40
        breq    NDIS45
        sbi     PORTD,6
        rjmp    NDIS46
NDIS43: cbi     PORTD,4
        rjmp    NDIS41
NDIS44: cbi     PORTD,5
        rjmp    NDIS42
NDIS45: cbi     PORTD,6

NDIS46: lds     Tmp1,DCFSTS
        cpi     Tmp1,4
        brne    NDIS47
        sbi     PORTD,7         ;Trennpunkt ein
NDIS47: rcall   W1ms            ;Warte 1ms
        cbi     PORTD,7         ;Trennpunkt aus
        sbi     PORTB,2         ;Vierte Stelle ausschalten
        lds     Tmp1,ESFLAG     ;Stromsparmodus?
        cpi     Tmp1,1
        brne    NDISPC
        rcall   W1ms            ;Warte 1ms
        rcall   W1ms            ;Warte 1ms
        rcall   W1ms            ;Warte 1ms

	;Zaehler ueberpruefen
NDISPC:	dec	r18	        ;Zaehler dekrementieren bis Schleife 250-mal
        breq	NDISPR		;durchlaufen
        rjmp    NDISP0 

NDISPR: pop     ZH
        pop     ZL
        pop     YH
        pop     YL
	pop	r18
        pop     r17
        pop     r16
        out     SREG,r16
        pop     r16
	ret			;Zurueck ins Hauptprogramm

;-------------------------------------------------------------------------------
;	Tabelle		(7-Segment-Dekodiertabelle (hoechstwertigstes Bit = 0))
;-------------------------------------------------------------------------------
NDISPT:
.DB	0x3F,0x06,0x5B,0x4F,0x66,0x6D,0x7D,0x07     ;Dekodiertabelle fuer
.DB	0x7F,0x6F,0x77,0x7C,0x39,0x5E,0x79,0x71     ;alle Hex-Zahlen