;*******************************************************************************
;*
;*	Titel:	wecker (wecker.asm)
;*	Datum:	15/08/09	Version:	0.2 (25/11/14)
;*	Autor:	WEIGU
;*
;*	Informationen zur Beschaltung:
;*      
;*	Prozessor: ATmega8			Quarzfrequenz: 8MHz
;*	Eingaenge:
;*	Ausgaenge:
;*
;*	Informationen zur Funktionsweise:
;*      
;*      Aenderung 25/11/14: Zeilen 309-316 schnellere Reaktion der Taster im 
;*      Nachtmodus
;*			
;*******************************************************************************

;-------------------------------------------------------------------------------
;	Einbinden der controllerspezifischen Definitionsdatei
;-------------------------------------------------------------------------------
.NOLIST				;List-Output ausschalten
.INCLUDE "m8def.inc"		;AVR-Definitionsdatei einbinden
.LIST				;List-Output wieder einschalten

;-------------------------------------------------------------------------------
;       Organisation des Datenspeichers (SRAM)
;-------------------------------------------------------------------------------
.DSEG                  ;was ab hier folgt kommt in den SRAM-Speicher
C_10ms: .BYTE  1        ;Ueberlauf-Zaehler für Timer2
SECPNT: .BYTE  1        ;Zeiger auf empfangene Sekunde
TSTFLG: .BYTE  1        ;Test Flag
DCFBIT: .BYTE  1        ;bei 100ms = 0 bei 200ms = 1
DCFERR: .BYTE  1        ;Zaehlt fehlerhaft empf. Bits
FR_END: .BYTE  1        ;Ende des Rahmens: 1 in 57s (0 in der 59 Sek.)
FR_PNT: .BYTE  1        ;Zeiger auf momentanen Rahmen
FINFNR: .BYTE  1        ;Anzahl der ganz durchgelaufenen Rahmen
FR_STS: .BYTE  1        ;Rahmenstatus
DCFSTS: .BYTE  1        ;DCF77 Status
MiBuf1: .BYTE  1        ;DCF Minuten-Zwischenspeicher 1 (zum Testen)
HrBuf1: .BYTE  1        ;DCF Stunden-Zwischenspeicher 1 (zum Testen)
MiBuf2: .BYTE  1        ;DCF Minuten-Zwischenspeicher 2 (zum Testen)
HrBuf2: .BYTE  1        ;DCF Stunden-Zwischenspeicher 2 (zum Testen)
AL_MIN: .BYTE  1        ;Weckzeit Minuten
AL_HR:  .BYTE  1        ;Weckzeit Stunden
ALFLAG: .BYTE  1        ;Flag ob Alarm gesetzt
ESFLAG: .BYTE  1        ;Flag ob gedimmtes Display (Nachts)
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
;       Programmspeicher (FLASH)     Programmstart nach RESET ab Adr. 0x0000
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.CSEG                           ;was ab hier folgt kommt in den FLASH-Speicher
.ORG	0x0000			;Programm beginnt an der FLASH-Adresse 0x0000
RESET1:	rjmp	INIT		;springe nach INIT (ueberspringe ISR Vektoren)

;-------------------------------------------------------------------------------
;	Sprungadressen fuer die Interrupts organisieren (ISR VECTORS)	
;-------------------------------------------------------------------------------
        ;Vektortabelle (im Flash-Speicher)
.ORG    INT1addr                ;interner Vektor fuer INT1
        rjmp    DCFI            ;Springe zur ISR von INT1 (DCFI)
.ORG    OVF2addr                ;interner Vektor fuer Timer2 Overflow
        rjmp    ISR_T2          ;Springe zur ISR von Timer2
.ORG    OVF0addr                ;interner Vektor fuer Timer0 Overflow
        rjmp    ISR_T0          ;Springe zur ISR von Timer0

;-------------------------------------------------------------------------------
;	Initialisierungen und eigene Definitionen
;-------------------------------------------------------------------------------
.ORG    INT_VECTORS_SIZE        ;Platz fuer ISR Vektoren lassen
INIT:
.DEF	Tmp1 = r16		;Register 16 dient als erster Zwischenspeicher
.DEF	Tmp2 = r17		;Register 17 dient als zweiter Zwischenspeicher

.DEF    Minutes = r18           ;Minutenregister
.DEF    Hours = r19             ;Stundenregister
.DEF    PTime0 = r20            ;Zeit während der Taster 0 betaetigt wurde
.DEF    PTime1 = r21            ;Zeit während der Taster 1 betaetigt wurde
.DEF    MinDCF = r22            ;DCF Minutenregister
.DEF    HrsDCF = r23            ;DCF Stundenregister
.DEF    WL = r24                ;Register 24 und 25 dienen als universelles
.DEF    WH = r25                ;Doppelregister W

        ;Stapel initialisieren (fuer Unterprogramme bzw. Interrupts)
        ldi	Tmp1,HIGH(RAMEND)	;RAMEND (SRAM) ist in der Definitions-
	out	SPH,Tmp1		;datei festgelegt
	ldi	Tmp1,LOW(RAMEND)
	out	SPL,Tmp1



        ;8 Pins der Segmentansteuerung als Ausgang initialisieren
        ldi     Tmp1,0x0F       ;4 untere Bits im Datenrichtungsregister auf 1
        out     DDRC,Tmp1       ;DDRC = 0b00001111
        ldi     Tmp1,0xF0       ;4 obere Bits im Datenrichtungsregister auf Eins
        out     DDRD,Tmp1       ;DDRC = 0b11110000
        ;4 Pins der Stellenansteuerung als Ausgang initialisieren
        ldi     Tmp1,0x3C       ;vier Bits im Datenrichtungsregister auf Eins
        out     DDRB,Tmp1       ;DDRB = 0b00111100

        ;2 Pins fuer Summer und PON (DCF-Modul einschalten) auf Ausgang
        sbi     DDRD,1          ;Summer
        sbi     DDRD,2          ;PON

        ;2 Pins fuer 2 Taster auf Eingang
        cbi     DDRB,0          ;
        cbi     DDRB,1          ;
        sbi     PORTB,0         ;Interne Pull-Ups
        sbi     PORTB,1         ;

        rcall   INII2C          ;I2C initialisieren

        ;Zaehler initialisieren
        clr     YL              ;Zaehler auf Null setzen
        clr     YH
        clr     MinDCF          ;loesche DCF Minutenregister
        clr     HrsDCF          ;loesche DCF Stundenregister
        clr     PTime0
        clr     PTime1

        ;Default Alarmzeit setzen und Alarmflag loeschen
        ldi     Tmp1,0x55
        sts     AL_MIN,Tmp1
        ldi     Tmp1,0x05
        sts     AL_HR,Tmp1
        clr     Tmp1
        sts     ALFLAG,Tmp1

        sei                       ;globales Interrupt-Flag setzen (Int.erlauben)
;-------------------------------------------------------------------------------
;	Hauptprogramm
;-------------------------------------------------------------------------------
MAIN:   cbi     PORTC,0         ;Segmente aus
        cbi     PORTC,1
        cbi     PORTC,2
        cbi     PORTC,3
        cbi     PORTD,4
        cbi     PORTD,5
        cbi     PORTD,6
        cbi     PORTB,5         ;erste Stelle einschalten für DCF LED
        sbi     PORTB,2         ;andere aus
        sbi     PORTB,3
        sbi     PORTB,4
        rcall   DCFON           ;DCF-Empfang einschalten (Interrupts init.)
W8DCF:  lds     Tmp1,DCFSTS
        cpi     Tmp1,1
        breq    LED2
        cpi     Tmp1,2
        breq    LED3
        cpi     Tmp1,3
        breq    WR_RTC
        cpi     Tmp1,4
        breq    MLOOP
        rjmp    W8DCF
LED2:   cbi     PORTB,4         ;vierte Stelle einschalten für DCF LED
        rjmp    W8DCF
LED3:   cbi     PORTB,3         ;zweite Stelle einschalten für DCF LED
        rjmp    W8DCF
WR_RTC: cbi     PORTD,7         ;DCF LEDs aus
        rcall   SNDALL          ;Daten zum RTC uebertragen
MLOOP:  rcall   RECALL          ;Daten vom RTC abholen
        cpi     Hours,0x04      ;Morgens 4 Uhr, DCF einschalten
        brne    ENESV
        cpi     Minutes,0x00
        brne    ENESV
        rjmp    MAIN
ENESV:  cpi     Hours,0x22      ;Stromsparmodus nachts aktivieren
        brge    ENESV1
        cpi     Hours,0x6       
        brlo    ENESV1
        clr     Tmp1
        sts     ESFLAG,Tmp1
        rjmp    TSTALF
ENESV1: ldi     Tmp1,1
        sts     ESFLAG,Tmp1
TSTALF: lds     Tmp1,ALFLAG     ;Alarmzeit erreicht?
        cpi     Tmp1,1
        brne    DISP
        lds     Tmp1,AL_HR      
        lds     Tmp2,AL_MIN
        cp      Tmp1,Hours
        brne    DISP
        cp      Tmp2,Minutes
        brne    DISP
        rcall   ALARM
DISP:   mov     YL,Minutes
        mov     YH,Hours
        rcall   NDISP           ;Darstellungsunterprogramm aufrufen
        clr     PTime0
        clr     PTime1
        rcall   DEBPB
        cpi     PTime0,0        ;beide Taster nicht gedrückt..
        brne    GO1
        cpi     PTime1,0
        breq    MLOOP           ;..weiter in Endlosschleife
GO1:    cpi     PTime0,15       ;beide Taster gleichzeitig laenger als 2s dann..
        brlo    GO2
        cpi     PTime1,15
        brlo    GO2            
        rjmp    MAIN            ;..DCF-Empfang manuell ausloesen
GO2:    cpi     PTime0,15       ;erster Taster laenger als 2s..
        brsh    SET_AL          ;Alarm ein bzw aus
GO3:    cpi     PTime1,15       ;zweiter Taster laenger als 2s..
        brsh    ST_ALT          ;Alarmzeit setzen
        cpi     PTime0,0        ;ein Taster kuerzer als 2s..
        brsh    SHOWAL
        cpi     PTime1,0        
        brsh    SHOWAL          ;;Alarmzeit zeigen
        rjmp    MLOOP

SHOWAL: lds     YL,AL_MIN
        lds     YH,AL_HR
        rcall   NDISP           ;Darstellungsunterprogramm aufrufen (1s)
        rcall   NDISP   
        rcall   NDISP
        rjmp    MLOOP           ;Weiter

SET_AL: lds     Tmp1, ALFLAG
        cpi     Tmp1,1
        breq    CLR_AL
        ldi     Tmp1,1
        sts     ALFLAG,Tmp1
        rjmp    GO3
CLR_AL: clr     Tmp1
        sts     ALFLAG,Tmp1
        rjmp    GO3

ST_ALT: rcall   SETALT
        rjmp    SHOWAL          ;Weiter

;-------------------------------------------------------------------------------
;	Unterprogramme und Interrupt-Behandlungsroutinen
;-------------------------------------------------------------------------------
; Interrupt-Behandlungsroutine von INT1 (DCF-Signal)
        ;Voreinstellung des Timer2 Zaehlregister auf 178 -> 78 Schritte
        ;Teiler = 1024 (CS20=CS21=CS22=1) (8MHz(Quarz)/1024 = 7812,5)
        ;7812/78 = 100/s -> Timerüberlauf nach 10ms
        ;Es wird nicht die Absenkung (100ms bzw. 200ms) erfasst, sondern der
        ; Traeger: 800ms = 1; 900ms = 0; 1800 bzw 1900ms = 59 Minute
        ;(58 Minute (Parität) wird ignoriert!) 
DCFI:   push    Tmp1            ;benutzte Reg. retten (Tmp1 = Zwischenspeicher)
        in      Tmp1,SREG       ;Statusregister einlesen
        push    Tmp1            ;Statusregister retten

        sbic    PIND,3          ;Ueberpruefe ob 1 (st. Fl.) oder 0 (fall. Fl.)
        rjmp    DCFI1
        in      Tmp1,TIMSK      ;fallende Flanke: starte den Timer
        ori     Tmp1,0x40       ;erlaube Timer2 Ueberlauf-Interrupt
        out     TIMSK,Tmp1
        cbi     PORTD,7         ;schalte LED ein zum Testen
        clr     Tmp1
        sts     C_10ms,Tmp1     ;Ueberlauf-Zaehler von Timer2 auf Null
        ldi     Tmp1,178        ;Timer2 Zaehlregister auf 178 (10ms mit 8MHz)
        out     TCNT2,Tmp1
        ldi     Tmp1,0x07       ;Timer einschalten (Teiler 1024)
        out     TCCR2,Tmp1
        rjmp    DCFIR           ;Raus
DCFI1:  in      Tmp1,TIMSK      ;steigende Flanke: stoppe den Timer
        andi    Tmp1,0xBF       ;sperre Timer2 Ueberlauf-Interrupt
        out     TIMSK,Tmp1
        sbi     PORTD,7         ;schalte LED aus zum Testen
        clr     Tmp1            ;Timer ausschalten
        out     TCCR2,Tmp1
        rcall   DCODE           ;Dekodiere
        rcall   TSTRS           ;Ueberpruefe Signal
DCFIR:  pop     Tmp1            ;Werte der geretteten Register wiederherstellen
        out     SREG,Tmp1        
        pop     Tmp1
        reti                    ;Rücksprung ins Hauptprogramm 

; Timer2-Interrupt-Routine fuer DCF77-Signal (Timer2 Overflow)
ISR_T2: push    Tmp1            ;benutzte Reg. retten (Tmp1 = Zwischenspeicher)
        in      Tmp1,SREG       ;Statusregister einlesen
        push    Tmp1            ;Statusregister retten
        ldi     Tmp1,178        ;Timer2 Zaehlregister auf 178 (10ms mit 8MHz)
        out     TCNT2,Tmp1
        lds     Tmp1,C_10ms
        inc     Tmp1            ;erhoehe den Ueberlauf-Zaehler von Timer2
        sts     C_10ms,Tmp1
        pop     Tmp1            ;Werte der geretteten Register wiederherstellen
        out     SREG,Tmp1
        pop     Tmp1
        reti                    ;Rücksprung ins Hauptprogramm

; Timer0-Interrupt-Routine fuer Alarm
ISR_T0: push    r16
        ldi     r16,100
        out     TCNT0,r16       ;Startwert festlegen
        in      r16,SREG        ;SREG-Register (Flags) retten
        push    r16
        ;Toggle Zustand von PD1 (hier ohne eor)
        sbic    PIND,1
        rjmp    ISRT00
        sbi     PORTD,1
        rjmp    ISRT0R
ISRT00: cbi     PORTD,1
ISRT0R: pop     r16             ;SREG-Register (Flags) wiederherstellen
        out     SREG,r16
        pop     r16
        reti                    ;zurueck und Interrupts wieder freigeben

; Zaehlprogramm fuer schliessende Taster (closer).
DEBPB:  push    r16
        clr     PTime0
        clr     PTime1
        in      r16,PINB
        com     r16
        andi    r16,0x03
        breq    DEBPBR
DEBPBT: lds     Tmp1,ESFLAG     ;Stromsparmodus?
        cpi     Tmp1,1
        breq    DEBPBT1
        call    W100ms          ;Zaehlzeit bei Tag
        rjmp    DEBPBT2
DEBPBT1: call    W10ms          ;Zaehlzeit bei Nacht (4* geringer)       
        call    W10ms          
DEBPBT2: cpi     r16,0x03
        breq    DEBPBB
        cpi     r16,0x01
        breq    DEBPB0
        inc     PTime1
        rjmp    DEBPBL
DEBPB0: inc     PTime0
        rjmp    DEBPBL
DEBPBB: inc     PTime0
        inc     PTime1
DEBPBL: in      r16,PINB
        com     r16
        andi    r16,0x03
        brne    DEBPBT
DEBPBR: pop     r16
        ret

; UP zum Setzen der Alarmzeit
SETALT:  push    r16
SETAL0: lds     YL,AL_MIN
        lds     YH,AL_HR
        rcall   NDISP           ;Darstellungsunterprogramm aufrufen
        sbic    PINB,0          ;Stunden setzen
        rjmp    STALH3          ;2. Taste gedrueckt dann um 1 erhoehen (STALH3)
        ;rcall   W100ms
        sbis    PINB,1          
        rjmp    STALM1          ;beide Tasten gedrueckt: Minuten setzen
        dec     YH              ;1. Taste gedrueckt dann um 1 erniedrigen
        cpi     YH,0xFF         ;Korrektur bei 0x00-1 = 0xFF (siehe STALH1)
        breq    STALH1
        mov     r16,YH
        andi    r16,0x0F        ;Korrektur bei 0x0-1 = 0xF (Subtr. von 6)
        cpi     r16,0x0F
        brne    STALH2
        ldi     r16,6
        sub     YH,r16
        rjmp    STALH2
STALH1: ldi     YH,0x23
STALH2: sts     AL_HR,YH        ;Alarmwert abspeichern
        rjmp    SETAL0
STALH3: sbic    PINB,1          ;2. Taste gedrueckt dann um 1 erhoehen
        rjmp    SETAL0
        inc     YH
        cpi     YH,0x24         ;Korrektur bei 0x23+1 = 0x24 (siehe STALH4)
        breq    STALH4
        mov     r16,YH
        andi    r16,0x0F        ;Korrektur bei 0x09+1 = 0xA (Add. von 6)
        cpi     r16,0x0A
        brne    STALH5
        ldi     r16,6
        add     YH,r16
        rjmp    STALH5
STALH4: clr     YH
STALH5: sts     AL_HR,YH        ;Alarmwert abspeichern
        rjmp    SETAL0
STALM1: rcall   NDISP           ;Warte 1s
        rcall   NDISP
        rcall   NDISP
        rcall   NDISP
        rcall   NDISP
STALM2: lds     YL,AL_MIN
        lds     YH,AL_HR
        rcall   NDISP           ;Darstellungsunterprogramm aufrufen
        sbic    PINB,0          ;Minuten setzen
        rjmp    STALM5          ;2. Taste gedrueckt dann um 1 erhoehen (STALM5)
        ;rcall   W10ms
        sbis    PINB,1
        rjmp    SETALR          ;beide Tasten gedrueckt: raus
        dec     YL              ;1. Taste gedrueckt dann um 1 erniedrigen
        cpi     YL,0xFF         ;Korrektur bei 0x00-1 = 0xFF (siehe STALH1)
        breq    STALM3
        mov     r16,YL
        andi    r16,0x0F        ;Korrektur bei 0x0-1 = 0xF (Subtr. von 6)
        cpi     r16,0x0F
        brne    STALM4
        ldi     r16,6
        sub     YL,r16
        rjmp    STALM4
STALM3: ldi     YL,0x59
STALM4: sts     AL_MIN,YL       ;Alarmwert abspeichern
        rjmp    STALM2
STALM5: sbic    PINB,1          ;2. Taste gedrueckt dann um 1 erhoehen
        rjmp    STALM2
        inc     YL
        cpi     YL,0x5A         ;Korrektur bei 0x59+1 = 0x60 (siehe STALM6)
        breq    STALM6
        mov     r16,YL
        andi    r16,0x0F        ;Korrektur bei 0x09+1 = 0xA (Add. von 6)
        cpi     r16,0x0A
        brne    STALM7
        ldi     r16,6
        add     YL,r16
        rjmp    STALM7
STALM6: clr     YL
STALM7: sts     AL_MIN,YL        ;Alarmwert abspeichern
        rjmp    STALM2
SETALR: pop     r16
        ret

; Alarm
ALARM:  push    r16
        push    r17
        ldi     r16,0x03        ;Timer0 einschalten und Vorteiler festlegen
        out     TCCR0,r16       ;TCCR0 = 0b00000100 (Systemtakt/128)
        in      r16,TIMSK       ;Interrupt Timer0 einschalten und freigeben
        ori     r16,0x01        ;Mit einer ODER-Maskierung (00000001b)                                
        out     TIMSK,r16       ;Bit 0 (TOIE0) im TIMSK-Register auf eins setzen
        sei                     ;Interrupts global freigeben
ALARM0: rcall   RECALL          ;Daten vom RTC abholen
        mov     YL,Minutes
        mov     YH,Hours
        rcall   NDISP           ;Darstellungsunterprogramm aufrufen
        sbis    PINB,0          ;
        rjmp    ALARM1
        sbic    PINB,1
        rjmp    ALARM0
ALARM1: clr     r16             ;Timer0 ausschalten
        out     TCCR0,r16       ;TCCR0 = 0
        in      r16,TIMSK       ;Interrupt Timer0 ausschalten
        andi    r16,0xFE        ;mit einer UND-Maskierung (11111110b)
        out     TIMSK,r16       ;Bit 0 (TOIE0) im TIMSK-Register auf null setzen
        cli                     ;Interrupts global sperren
ALARM2: rcall   RECALL          ;Daten vom RTC abholen
        mov     YL,Minutes
        mov     YH,Hours
        rcall   NDISP           ;Darstellungsunterprogramm aufrufen
        lds     Tmp1,AL_HR      ;Warten bis Alarmzeit und Uhrzeit nicht mehr
        lds     Tmp2,AL_MIN     ;gleich!
        cp      Tmp1,Hours
        brne    ALARMR
        cp      Tmp2,Minutes
        brne    ALARMR
        rjmp    ALARM2
ALARMR: pop     r17
        pop     r16
        ret

.INCLUDE "SR_TIME_8M.asm"       ;Zeitschleifenbibliothek einbinden
.INCLUDE "SR_DCF77.asm"         ;DCF77-Bibliothek einbinden
.INCLUDE "SR_NUMDISPLAY_W.asm"  ;7Seg-Anzeigen UP einbinden
.INCLUDE "SR_I2C_W.asm"         ;I2C (RTC) UP einbinden
;-------------------------------------------------------------------------------
;       Tabellen im Programmspeicher (Flash)
;-------------------------------------------------------------------------------
;TAB:  .DB    "Hallo"

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.EXIT				;Ende des Quelltextes
